// window.cpp

#include <map>
#include <ifmwindow>

#pragma warning(disable: 4786)

#ifdef DIRECT
#  define BUFFERED false
#else
#  define BUFFERED true
#endif

using namespace std;

static map<int, IfmWindow*> allWindows;      // mapping window id to pointer
IfmWindow wio;                               // The default window

// --------------------------------------------------------------------
//                        IfmWindow functions, part 1 (Win32 independent)
// --------------------------------------------------------------------


bool IfmWindow::check_key()         { pollEvent(); return key.pressed;    }
bool IfmWindow::check_mouse()       { pollEvent(); return mouse.motion;   }
bool IfmWindow::check_mouse_click() { pollEvent(); return mouse.released; }
 
int IfmWindow::get_key()
{
    while (!check_key());
    key.pressed = false;
    return key.ascii;
}

void IfmWindow::get_mouse(int& x, int& y)
{
    while (!check_mouse());
    x = mouse.x;
    y = mouse.y;
    mouse.motion = false;
}

int IfmWindow::get_mouse_click(int& x, int& y)
{
    while (!check_mouse_click());
    x = mouse.x;
    y = mouse.y;
    mouse.released = false;
    return mouse.button;
}

void IfmWindow::wait_for_mouse_click(int i)
{
    int x, y;
    while (get_mouse_click(x, y) != i && i != 0);  // 0: any button
}

// ----- input methods:

void Point::input(IfmWindow& win) 
{
    win.makeVisible();
    win.get_mouse_click(_x, _y);
}

void PointPair::input(IfmWindow& win)
{
    Pen save = win.pen;

    win.makeVisible();

    win.pen.mode  = GXxor;
    win.pen.color = white.index;
    win.setBuffered(false);             // draw immediately while in input mode

    win.get_mouse_click(_x1, _y1);
    _x2 = _x1;  _y2 = _y1;

    bool done = false;
    do {
        draw(win);                      // draw preview ...
        while (!win.check_mouse() && !win.check_mouse_click());
        int x, y;
        if (win.check_mouse()) win.get_mouse(x, y);
        else { done = true;    win.get_mouse_click(x, y); }
        draw(win);                      // and erase it
        _x2 = x;  _y2 = y;
    } while (!done);

    win.setBuffered(BUFFERED);

    win.pen = save;
}

// ----- Functions needed by manipulators:
IfmWindow& sync      (IfmWindow& w) { w.mouse.released = w.key.pressed = 
                                      w.mouse.motion = false;      return w; }
IfmWindow& endl      (IfmWindow& w)                    { sync(w);  return w; }
IfmWindow& line_width(IfmWindow& w, int i) { w.set_line_width(i);  return w; }
IfmWindow& draw_mode (IfmWindow& w, int i) { w.set_draw_mode (i);  return w; }
IfmWindow& copy_mode (IfmWindow& w)    { w.set_draw_mode(GXcopy);  return w; }
IfmWindow&  xor_mode (IfmWindow& w)    { w.set_draw_mode(GXxor );  return w; }
IfmWindow&   or_mode (IfmWindow& w)    { w.set_draw_mode(GXor  );  return w; }
IfmWindow&  and_mode (IfmWindow& w)    { w.set_draw_mode(GXand );  return w; }

// ----- 1-arg manipulators: 
ManipU wait(ulong u)      { return ManipU(&wait      , u); }
ManipI line_width(int i)  { return ManipI(&line_width, i); }
ManipI draw_mode(int i)   { return ManipI(&draw_mode , i); }

// ----- I/O operators:
IfmWindow& operator<<(IfmWindow& w, const Shape& s) { s.draw(w);   return w; }
IfmWindow& operator<<(IfmWindow& w, const color& c) { w.set_color(c.index); 
                                                                   return w; }
IfmWindow& operator<<(IfmWindow& w, IfmWindow& (*f)(IfmWindow&)) { return 
                                                                   (*f)(w);  }
IfmWindow& operator<<(IfmWindow& w, ManipI m) { (m.fun)(w, m.arg); return w; }
IfmWindow& operator<<(IfmWindow& w, ManipU m) { (m.fun)(w, m.arg); return w; }

IfmWindow& operator>>(IfmWindow& w, Shape& s) { s.input(w);        return w; }


// --------------------------------------------------------------------
//                        Start of Win32 specific part
// --------------------------------------------------------------------

#include <windows.h>

COLORREF colormap[27] = {  // Format:  bbggrr
    0x0000ff, 0x0020ff, 0x0040ff, 0x0060ff,         // red      ...
    0x0080ff, 0x00a0ff, 0x00c0ff, 0x00e0ff,         // orange   ...
    0x00ffff, 0x00eec0, 0x00dd80, 0x00cc40,         // yellow   ...
    0x00bb00, 0x409900, 0x807700, 0xc05500,         // green    ...
    0xff3300, 0xff2233, 0xff1166, 0xff0099,         // blue     ...
    0xff00cc, 0xdd00dd, 0xbb00ee, 0x9900ff,         // magenta  ...
    0x7700ff,                                       // purple
    0x000000,                                       // black
    0xffffff                                        // white
};

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

void IfmWindow::makeVisible() { 
    if (!isVisible) {
        ShowWindow((HWND)id, SW_SHOW); isVisible = true; 
    }
}

// --------------------------------------------------------------------
//                        IfmWindow functions, part 2 (Win32 specific)
// --------------------------------------------------------------------

void IfmWindow::create(int xpos, int ypos, int xsize, int ysize, string str)
{
    width          = xsize;
    height         = ysize;
    offLeft        = xpos;
    offTop         = ypos;
    title          = str;
    pen.width      = 1;
    pen.mode       = GXcopy;
    pen.color      = black.index;
    mouse.motion   = false;
    mouse.released = false;
    key.pressed    = false;
    isVisible      = false;

                // Define and register a window class
    WNDCLASS wc      = { 0 };
    wc.style         = CS_OWNDC;
    wc.lpfnWndProc   = WndProc;
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wc.lpszClassName = "Foo";
    RegisterClass(&wc);

                // Create the window
    HWND hwnd = CreateWindowEx(WS_EX_TOPMOST, "Foo", title.c_str(), 
                               WS_OVERLAPPEDWINDOW, offLeft, offTop, 
                               3+width+3, 27+height+3, NULL, NULL, NULL, NULL);

    HDC hdc  = GetDC(hwnd);                  // Get the device context
    HDC hdc1 = CreateCompatibleDC(hdc);      // Create a memory DC
    SelectObject(hdc1, CreateCompatibleBitmap(hdc, width, height));

    id     = (int)hwnd;                      // Handle for window
    screen = (int)hdc;                       // Handle for screen DC
    memory = (int)hdc1;                      // Handle for memory DC

    allWindows[id] = this;        
    buffered       = BUFFERED;

    clear();
}

IfmWindow::~IfmWindow()        // Destructor
{
    if (id != 0) {
        HWND hwnd = (HWND)id;
        DeleteDC     ((HDC)memory);
        DeleteDC     ((HDC)screen);
        DestroyWindow(hwnd);
        if (allWindows.size() != 0) allWindows.erase(allWindows.find(id));
    }
}

// ----- Some auxiliary functions

void IfmWindow::pollEvent()
{
    MSG msg;

    makeVisible();
    if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        TranslateMessage(&msg);
        DispatchMessage (&msg);
    }
}

// ----- Functions needed by manipulators

IfmWindow& wait(IfmWindow& w, ulong us) 
{ 
    Sleep(us/1000);  
    return w; 
}

IfmWindow& flush(IfmWindow& w)                                  
{
    if (w.buffered) {
        w.makeVisible();
                          // Copy bitmap from memory to screen
        BitBlt((HDC)w.screen, 0, 0, w.width-1, w.height-1,
               (HDC)w.memory, 0, 0, SRCCOPY);
    }
    return w;
}

IfmWindow& clear(IfmWindow& w)
{
    if (w.buffered) PatBlt((HDC)w.memory, 0, 0, w.width, w.height, WHITENESS);
    PatBlt((HDC)w.screen, 0, 0, w.width, w.height, WHITENESS);
    return w;
}

// --------------------------------------------------------------------
//                                Shape functions (Win32 specific)
// --------------------------------------------------------------------


HDC prepareDraw(IfmWindow& win, bool line, bool filled) 
{
    win.makeVisible();

    HDC hdc = (HDC)win.device();

    switch(win.pen.mode) {
        case GXcopy: SetROP2(hdc, R2_COPYPEN);  break;
        case GXxor : SetROP2(hdc, R2_XORPEN);   break;
        case GXor  : SetROP2(hdc, R2_MERGEPEN); break;
        case GXand : SetROP2(hdc, R2_MASKPEN);  break;
    }

    if (line) {
        LOGBRUSH lb = { BS_NULL };
        DeleteObject(SelectObject(hdc, CreateBrushIndirect(&lb)));
        DeleteObject(SelectObject(hdc, CreatePen(PS_SOLID, win.pen.width, 
                                                 colormap[win.pen.color])));
    }
    else if (filled) {
        LOGBRUSH lb = { BS_SOLID, colormap[win.pen.color] };
        DeleteObject(SelectObject(hdc, CreateBrushIndirect(&lb)));
        DeleteObject(SelectObject(hdc, CreatePen(PS_NULL, 0, 0)));
    }
    return hdc;
}

// ----- drawing methods:

void Point::draw(IfmWindow& win) const
{
    if (win.pen.width == 1) {
        HDC hdc = prepareDraw(win, false, false);

        SetPixel(hdc, _x, win.vflip(_y), colormap[win.pen.color]);
    }
    else {
        HDC hdc = prepareDraw(win, true, false);

        MoveToEx(hdc, _x, win.vflip(_y), (LPPOINT)NULL);
        LineTo  (hdc, _x, win.vflip(_y));
    }
}

void Line::draw(IfmWindow& win) const
{
    HDC hdc = prepareDraw(win, true, false);

    MoveToEx(hdc, _x1, win.vflip(_y1), (LPPOINT)NULL);
    LineTo  (hdc, _x2, win.vflip(_y2));
}

void Rectangle::draw(IfmWindow& win)  const
{
    HDC hdc = prepareDraw(win, !filled, filled);

    ::Rectangle(hdc, _x1, win.vflip(_y1), _x2, win.vflip(_y2));
}

void Ellipse::draw(IfmWindow& win) const
{
    HDC hdc = prepareDraw(win, !filled, filled);

    int dx = w(), dy = h();

    if (circle) { if (dx < dy) dx = dy; else dy = dx; }

    ::Ellipse(hdc, x()-dx, win.vflip(y()+dy), x()+dx, win.vflip(y()-dy));
}

// --------------------------------------------------------------------
//                                The event handler (Win32 specific)
// --------------------------------------------------------------------

void debugEvent(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{ 
    //debugEvent(hwnd, message, wParam, lParam);

    if (allWindows.size()) {
        IfmWindow* ptr = allWindows[(int)hwnd];
        if (ptr) {
            IfmWindow& w = *ptr;

            switch (message) {
                case WM_MOUSEMOVE: 
                    w.mouse.x        = LOWORD(lParam);
                    w.mouse.y        = w.vflip(HIWORD(lParam));
                    w.mouse.motion   = true;
                    break;
                case WM_LBUTTONUP:   
                    w.mouse.released = true;
                    w.mouse.button   = 1;
                    break;
                case WM_MBUTTONUP:   
                    w.mouse.released = true;
                    w.mouse.button   = 2;
                    break;
                case WM_RBUTTONUP:   
                    w.mouse.released = true;
                    w.mouse.button   = 3;
                    break;
                case WM_KEYDOWN:   
                    w.key.pressed    = true; 
                    w.key.ascii      = wParam;
                    if (wParam == VK_ESCAPE) exit(0);
                    break;
                case WM_SIZE: 
                case WM_EXITSIZEMOVE: 
                case WM_ACTIVATE: 
                    flush(w);
                    break;
                case WM_CLOSE: 
                    exit(0);
                    break;
                default: 
                    break;
            }
        }
    }
            // finally, call the default event handler:
    return DefWindowProc(hwnd, message, wParam, lParam);
}


//---

#include <cstdio>
#include <cstring>
#define HFILE "/usr/include/w32api/winuser.h"

void debugEvent(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static char eventName[1024][41] = {""};
    static int ct[1024];
    static int ctTotal = 0;
    static bool firstCall = true;
    static bool ok = true;

    if (firstCall) {
	FILE* fp = fopen(HFILE, "r");
	char line[101];

	if (!fp) {
	    fprintf(stderr, "Can't find file %s", HFILE);
	    ok = false;
	}
	while (fgets(line, 100, fp)) {
	    char str1[100], str2[100];
	    unsigned int num;

	    if (3 != sscanf(line, "%s%s%d", str1, str2, &num)) continue;
	    if (num >= 1024) continue;
	    if (strcmp("#define", str1)) continue;
	    if (strncmp("WM_", str2, 3)) continue;
	    //if (strcmp(eventName[num], "")) continue;
	    strncpy(eventName[num], str2, 40);
	}
	for (int i = 0; i < 1024; i++) ct[i] = 0;

	firstCall = false;
    }
    if (!ok) return;

    int total = ctTotal++;
    int nr = (unsigned int)message;

    if (nr >= 1024) return;
    if ((ct[nr]++) > 20) return;

    fprintf(stderr, "%5d %-22s (win %03d)%6d%6d%12d\n",
        total,
    	eventName[nr],
	((int)hwnd) % 1000,
        (int)HIWORD(lParam),
        (int)LOWORD(lParam),
        (int)wParam
    );
}
